import asyncio
import logging
from typing import TYPE_CHECKING, Any

from typing_extensions import override

from hllrcon.commands import RconCommands
from hllrcon.exceptions import HLLConnectionLostError
from hllrcon.protocol.protocol import RconProtocol

if TYPE_CHECKING:
    from collections.abc import Callable


class RconConnection(RconCommands):
    """A class representing a connection to an RCON server.

    RconConnections are single-use and cannot be reused after being disconnected from
    the RCON server. For a RCON client that includes recovery mechanisms for connection
    issues, refer to ~`Rcon` instead.
    """

    def __init__(self, protocol: RconProtocol) -> None:
        self._protocol = protocol
        self._disconnect_event: asyncio.Event = asyncio.Event()
        self._disconnect_event.set()

        self.on_disconnect: Callable[[], None] = lambda: None

    def is_connected(self) -> bool:
        """Check if the connection is still active.

        Returns
        -------
        bool
            True if the connection is active, False otherwise.

        """
        return self._protocol.is_connected()

    def disconnect(self) -> None:
        """Disconnect from the RCON server."""
        self._protocol.disconnect()

    def _on_disconnect(self, _: Exception | None) -> None:
        """Internal callback for when the connection is lost."""
        self._disconnect_event.set()
        self.on_disconnect()

    async def wait_until_disconnected(self) -> None:
        """Wait until the connection is closed."""
        await self._disconnect_event.wait()

    @classmethod
    async def connect(
        cls,
        host: str,
        port: int,
        password: str,
        logger: logging.Logger | None = None,
    ) -> "RconConnection":
        """Connect to the RCON server.

        Parameters
        ----------
        host : str
            The hostname or IP address of the RCON server.
        port : int
            The port number of the RCON server.
        password : str
            The password for the RCON server.
        logger : logging.Logger | None, optional
            A logger instance for logging messages, by default None. If None,
            `logging.getLogger(__name__)` is used.

        Raises
        ------
        HLLConnectionError
            The address and port could not be resolved.
        HLLConnectionRefusedError
            The server refused the connection.
        HLLAuthError
            The provided password is incorrect.

        Returns
        -------
        RconConnection
            A connection to the RCON server.

        """
        protocol = await RconProtocol.connect(
            host=host,
            port=port,
            password=password,
            logger=logger,
        )
        self = cls(protocol)

        self._disconnect_event.clear()
        self._protocol.on_connection_lost = self._on_disconnect

        return self

    @override
    async def execute(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
    ) -> str:
        if self._disconnect_event.is_set():
            raise HLLConnectionLostError
        response = await self._protocol.execute(command, version, body)
        response.raise_for_status()
        return response.content_body



# ruff: noqa: N802

from enum import StrEnum

from hllrcon.data._utils import IndexedBaseModel, class_cached_property


class RoleType(StrEnum):
    INFANTRY = "Infantry"
    ARMOR = "Armor"
    RECON = "Recon"
    COMMANDER = "Commander"


class Role(IndexedBaseModel[int]):
    id: int
    name: str
    pretty_name: str
    type: RoleType
    is_squad_leader: bool
    """Whether this role is exclusive to the squad leader.

    Roles included are:
    - Commander
    - Officer
    - Tank Commander
    - Spotter
    """
    kill_combat_score: int
    """The Combat Effectiveness score gained when killing a player with this role."""
    assist_combat_score: int
    """The Combat Effectiveness score gained when assisting in the kill of a player with
    this role. Assists are gained when a player you downed gets killed by another."""

    @property
    def is_infantry(self) -> bool:
        """Whether the role is associated with infantry units.

        Roles included are:
        - Officer
        - Rifleman
        - Assault
        - Automatic Rifleman
        - Medic
        - Support
        - Machine Gunner
        - Anti-Tank
        - Engineer
        """
        return self.type == RoleType.INFANTRY

    @property
    def is_tanker(self) -> bool:
        """Whether the role is associated with armor units.

        Roles included are:
        - Tank Commander
        - Crewman
        """
        return self.type == RoleType.ARMOR

    @property
    def is_recon(self) -> bool:
        """Whether the role is associated with recon units.

        Roles included are:
        - Spotter
        - Sniper
        """
        return self.type == RoleType.RECON

    @class_cached_property
    @classmethod
    def RIFLEMAN(cls) -> "Role":
        return cls(
            id=0,
            name="Rifleman",
            pretty_name="Rifleman",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=3,
            assist_combat_score=2,
        )

    @class_cached_property
    @classmethod
    def ASSAULT(cls) -> "Role":
        return cls(
            id=1,
            name="Assault",
            pretty_name="Assault",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def AUTOMATIC_RIFLEMAN(cls) -> "Role":
        return cls(
            id=2,
            name="AutomaticRifleman",
            pretty_name="Automatic Rifleman",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def MEDIC(cls) -> "Role":
        return cls(
            id=3,
            name="Medic",
            pretty_name="Medic",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def SPOTTER(cls) -> "Role":
        return cls(
            id=4,
            name="Spotter",
            pretty_name="Spotter",
            type=RoleType.RECON,
            is_squad_leader=True,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def SUPPORT(cls) -> "Role":
        return cls(
            id=5,
            name="Support",
            pretty_name="Support",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def MACHINE_GUNNER(cls) -> "Role":
        return cls(
            id=6,
            name="HeavyMachineGunner",
            pretty_name="Machine Gunner",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=9,
            assist_combat_score=6,
        )

    @class_cached_property
    @classmethod
    def ANTI_TANK(cls) -> "Role":
        return cls(
            id=7,
            name="AntiTank",
            pretty_name="Anti-Tank",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=9,
            assist_combat_score=6,
        )

    @class_cached_property
    @classmethod
    def ENGINEER(cls) -> "Role":
        return cls(
            id=8,
            name="Engineer",
            pretty_name="Engineer",
            type=RoleType.INFANTRY,
            is_squad_leader=False,
            kill_combat_score=9,
            assist_combat_score=6,
        )

    @class_cached_property
    @classmethod
    def OFFICER(cls) -> "Role":
        return cls(
            id=9,
            name="Officer",
            pretty_name="Officer",
            type=RoleType.INFANTRY,
            is_squad_leader=True,
            kill_combat_score=9,
            assist_combat_score=6,
        )

    @class_cached_property
    @classmethod
    def SNIPER(cls) -> "Role":
        return cls(
            id=10,
            name="Sniper",
            pretty_name="Sniper",
            type=RoleType.RECON,
            is_squad_leader=False,
            kill_combat_score=6,
            assist_combat_score=4,
        )

    @class_cached_property
    @classmethod
    def CREWMAN(cls) -> "Role":
        return cls(
            id=11,
            name="Crewman",
            pretty_name="Crewman",
            type=RoleType.ARMOR,
            is_squad_leader=False,
            kill_combat_score=3,
            assist_combat_score=2,
        )

    @class_cached_property
    @classmethod
    def TANK_COMMANDER(cls) -> "Role":
        return cls(
            id=12,
            name="TankCommander",
            pretty_name="Tank Commander",
            type=RoleType.ARMOR,
            is_squad_leader=True,
            kill_combat_score=9,
            assist_combat_score=6,
        )

    @class_cached_property
    @classmethod
    def COMMANDER(cls) -> "Role":
        return cls(
            id=13,
            name="ArmyCommander",
            pretty_name="Army Commander",
            type=RoleType.COMMANDER,
            is_squad_leader=True,
            kill_combat_score=12,
            assist_combat_score=8,
        )



# ruff: noqa: N802


from ._utils import IndexedBaseModel, class_cached_property


class Team(IndexedBaseModel[int]):
    id: int
    name: str

    @class_cached_property
    @classmethod
    def ALLIES(cls) -> "Team":
        return cls(id=1, name="Allies")

    @class_cached_property
    @classmethod
    def AXIS(cls) -> "Team":
        return cls(id=2, name="Axis")



######################################################
#  Generated using `scripts/generate_sync_commands.py`
#  DO NOT EDIT THIS FILE DIRECTLY
######################################################

import json
from functools import partial
from typing import Any, Literal
from unittest import mock

import pytest
from hllrcon.exceptions import HLLCommandError, HLLMessageError
from hllrcon.responses import (
    ForceMode,
)
from hllrcon.sync.commands import (
    SyncRconCommands,
    cast_response_to_bool,
    cast_response_to_model,
)
from pydantic import BaseModel, ValidationError


class SyncRconCommandsStub(SyncRconCommands):
    def __init__(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
        response: str | HLLCommandError = "",
    ) -> None:
        super().__init__()
        self.command = command
        self.version = version
        self.body = body
        self.response = response

    def execute(
        self,
        command: str,
        version: int,
        body: str | dict[str, Any] = "",
    ) -> str:
        assert command == self.command
        assert version == self.version
        assert body == self.body

        if isinstance(self.response, HLLCommandError):
            raise self.response
        return self.response


class CustomModel(BaseModel):
    x: int
    y: int


class TestCastResponseToModel:
    def test_cast_response_to_model_success(self) -> None:
        called = {}

        @cast_response_to_model(CustomModel)
        def dummy_func(x: int, y: int = 0) -> str:
            called["x"] = x
            called["y"] = y
            return f'{{"x": {x + 1}, "y": {y + 1}}}'

        result = dummy_func(5, y=10)
        assert result.x == 6
        assert result.y == 11
        assert called == {"x": 5, "y": 10}

    def test_cast_response_to_model_raises_on_invalid_json(self) -> None:
        @cast_response_to_model(CustomModel)
        def dummy_func() -> str:
            return "not a json"

        with pytest.raises(ValidationError, match="Invalid JSON"):
            dummy_func()

    def test_cast_response_to_model_raises_on_non_dict(self) -> None:
        @cast_response_to_model(CustomModel)
        def dummy_func() -> str:
            return '["not a dict"]'

        with pytest.raises(ValidationError, match="Input should be an object"):
            dummy_func()

    def test_cast_response_to_model_preserves_function_metadata(self) -> None:
        @cast_response_to_model(CustomModel)
        def dummy_func() -> str:
            """Docstring here."""
            return '{"a": 1}'

        assert dummy_func.__name__ == "dummy_func"
        assert dummy_func.__doc__ == "Docstring here."


class TestCastResponseToBool:
    def test_cast_response_to_bool_success(self) -> None:
        called = {}

        @cast_response_to_bool({400})
        def dummy_func() -> None:
            called["called"] = True

        result = dummy_func()
        assert result is True
        assert called == {"called": True}

    def test_cast_response_to_bool_failure(self) -> None:
        called = {}

        @cast_response_to_bool({400})
        def dummy_func() -> None:
            called["called"] = True
            msg = "Command failed"
            raise HLLCommandError(400, msg)

        result = dummy_func()
        assert result is False
        assert called == {"called": True}

    def test_cast_response_to_bool_exception(self) -> None:
        called = {}
        msg = "Internal server error"

        @cast_response_to_bool({400})
        def dummy_func() -> None:
            called["called"] = True
            raise HLLCommandError(500, msg)

        with pytest.raises(HLLCommandError, match=msg):
            dummy_func()
        assert called == {"called": True}


class TestCommands:
    def test_commands_add_admin(self) -> None:
        player_id = "Player ID"
        admin_group = "Admin Group"
        comment = "Comment"

        SyncRconCommandsStub(
            "AddAdmin",
            2,
            {
                "PlayerId": player_id,
                "AdminGroup": admin_group,
                "Comment": comment,
            },
        ).add_admin(player_id, admin_group, comment)

    def test_commands_remove_admin(self) -> None:
        player_id = "Player ID"

        SyncRconCommandsStub(
            "RemoveAdmin",
            2,
            {"PlayerId": player_id},
        ).remove_admin(player_id)

    @pytest.mark.parametrize("filter_", ["Filter", None])
    def test_commands_admin_log(
        self,
        filter_: str | None,
    ) -> None:
        SyncRconCommandsStub(
            "GetAdminLog",
            2,
            {"LogBackTrackTime": 60, "Filters": filter_ or ""},
            (
                '{"entries": ['
                '{"timestamp": "2023-10-01T12:00:00Z", "message": "Log 1"},'
                '{"timestamp": "2023-10-01T12:01:00Z", "message": "Log 2"}'
                "]}"
            ),
        ).get_admin_log(60, filter_)

    def test_commands_admin_log_seconds_span_invalid(self) -> None:
        with pytest.raises(
            ValueError,
            match="seconds_span must be a non-negative integer",
        ):
            SyncRconCommandsStub(
                "GetAdminLog",
                2,
                {"LogBackTrackTime": -1, "Filters": ""},
            ).get_admin_log(-1)

    def test_commands_change_map(self) -> None:
        map_name = "Map Name"
        SyncRconCommandsStub(
            "ChangeMap",
            2,
            {"MapName": map_name},
        ).change_map(map_name)

    def test_commands_get_available_sector_names(self) -> None:
        command_id = "SetSectorLayout"
        sector_names: tuple[list[str], ...] = (
            ["ROAD TO RECOGNE", "COBRU APPROACH", "ROAD TO NOVILLE"],
            ["COBRU FACTORY", "FOY", "FLAK BATTERY"],
            ["WEST BEND", "SOUTHERN EDGE", "DUGOUT BARN"],
            ["N30 HIGHWAY", "BIZORY-FOY ROAD", "EASTERN OURTHE"],
            ["ROAD TO BASTOGNE", "BOIS JACQUES", "FOREST OUTSKIRTS"],
        )
        response = SyncRconCommandsStub(
            "GetClientReferenceData",
            2,
            command_id,
            json.dumps(
                {
                    "name": "SetSectorLayout",
                    "text": "Set Sector Layout",
                    "description": "Configure the active sector layout",
                    "dialogueParameters": [
                        {
                            "type": "Combo",
                            "name": f"Sector {i + 1}",
                            "iD": f"Sector_{i + 1}",
                            "displayMember": ",".join(sector_names[i]),
                            "valueMember": ",".join(sector_names[i]),
                        }
                        for i in range(5)
                    ],
                },
            ),
        ).get_available_sector_names()
        assert response == sector_names

    def test_commands_get_available_sector_names_invalid_message(self) -> None:
        command_id = "SetSectorLayout"
        stub = SyncRconCommandsStub(
            "GetClientReferenceData",
            2,
            command_id,
            json.dumps(
                {
                    "name": "SetSectorLayout",
                    "text": "Set Sector Layout",
                    "description": "Configure the active sector layout",
                    "dialogueParameters": [
                        {
                            "type": "Combo",
                            "name": "Sector 1",
                            "iD": "Some incorrect ID",
                            "displayMember": "blib,blab,blob",
                            "valueMember": "blib,blab,blob",
                        },
                    ],
                },
            ),
        )

        with pytest.raises(HLLMessageError):
            stub.get_available_sector_names()

    def test_commands_change_sector_layout(self) -> None:
        sector1 = "Sector 1"
        sector2 = "Sector 2"
        sector3 = "Sector 3"
        sector4 = "Sector 4"
        sector5 = "Sector 5"

        SyncRconCommandsStub(
            "SetSectorLayout",
            2,
            {
                "Sector_1": sector1,
                "Sector_2": sector2,
                "Sector_3": sector3,
                "Sector_4": sector4,
                "Sector_5": sector5,
            },
        ).set_sector_layout(sector1, sector2, sector3, sector4, sector5)

    def test_commands_add_map_to_rotation(self) -> None:
        map_name = "Map1"
        index = 3
        SyncRconCommandsStub(
            "AddMapToRotation",
            2,
            {"MapName": map_name, "Index": index},
        ).add_map_to_rotation(map_name, index)

    def test_commands_remove_map_from_rotation(self) -> None:
        index = 2
        SyncRconCommandsStub(
            "RemoveMapFromRotation",
            2,
            {"Index": index},
        ).remove_map_from_rotation(index)

    def test_commands_add_map_to_sequence(self) -> None:
        map_name = "Map2"
        index = 1
        SyncRconCommandsStub(
            "AddMapToSequence",
            2,
            {"MapName": map_name, "Index": index},
        ).add_map_to_sequence(map_name, index)

    def test_commands_remove_map_from_sequence(self) -> None:
        index = 4
        SyncRconCommandsStub(
            "RemoveMapFromSequence",
            2,
            {"Index": index},
        ).remove_map_from_sequence(index)

    def test_commands_set_map_shuffle_enabled(self) -> None:
        SyncRconCommandsStub(
            "SetMapShuffleEnabled",
            2,
            {"Enable": True},
        ).set_map_shuffle_enabled(enabled=True)

    def test_commands_move_map_from_sequence(self) -> None:
        old_index = 1
        new_index = 2
        SyncRconCommandsStub(
            "MoveMapInSequence",
            2,
            {"CurrentIndex": old_index, "NewIndex": new_index},
        ).move_map_in_sequence(old_index, new_index)

    def test_commands_get_available_maps(self) -> None:
        command_id = "AddMapToRotation"
        maps = ["foy_warfare", "stmariedumont_warfare", "hurtgenforest_warfare_V2"]
        response = SyncRconCommandsStub(
            "GetClientReferenceData",
            2,
            command_id,
            json.dumps(
                {
                    "name": command_id,
                    "text": "Add Map to Rotation",
                    "description": "Add a new map to the map rotation at an index.",
                    "dialogueParameters": [
                        {
                            "type": "Combo",
                            "name": "Map Name",
                            "iD": "MapName",
                            "displayMember": ",".join(maps),
                            "valueMember": ",".join(maps),
                        },
                        {
                            "type": "Number",
                            "name": "At Index",
                            "iD": "Index",
                            "displayMember": "",
                            "valueMember": "",
                        },
                    ],
                },
            ),
        ).get_available_maps()
        assert response == maps

    def test_commands_get_available_maps_invalid_message(self) -> None:
        command_id = "AddMapToRotation"
        maps = ["foy_warfare", "stmariedumont_warfare", "hurtgenforest_warfare_V2"]
        stub = SyncRconCommandsStub(
            "GetClientReferenceData",
            2,
            command_id,
            json.dumps(
                {
                    "name": command_id,
                    "text": "Add Map to Rotation",
                    "description": "Add a new map to the map rotation at an index.",
                    "dialogueParameters": [
                        {
                            "type": "Combo",
                            "name": "Map Name",
                            "iD": "NotMapName",  # Changed "MapName" to "NotMapName"
                            "displayMember": ",".join(maps),
                            "valueMember": ",".join(maps),
                        },
                        {
                            "type": "Number",
                            "name": "At Index",
                            "iD": "Index",
                            "displayMember": "",
                            "valueMember": "",
                        },
                    ],
                },
            ),
        )

        with pytest.raises(HLLMessageError):
            stub.get_available_maps()

    def test_commands_get_commands(self) -> None:
        SyncRconCommandsStub(
            "GetDisplayableCommands",
            2,
            response=json.dumps(
                {
                    "entries": [
                        {
                            "iD": "cmd1",
                            "friendlyName": "Command 1",
                            "isClientSupported": True,
                        },
                        {
                            "iD": "cmd2",
                            "friendlyName": "Command 2",
                            "isClientSupported": False,
                        },
                    ],
                },
            ),
        ).get_commands()

    def test_commands_get_admin_groups(self) -> None:
        SyncRconCommandsStub(
            "GetAdminGroups",
            2,
            response=json.dumps(
                {
                    "groupNames": [
                        "owner",
                        "senior",
                        "junior",
                        "spectator",
                    ],
                },
            ),
        ).get_admin_groups()

    def test_commands_get_admin_users(self) -> None:
        SyncRconCommandsStub(
            "GetAdminUsers",
            2,
            response=json.dumps(
                {
                    "adminUsers": [
                        {
                            "userId": "user1",
                            "group": "owner",
                            "comment": "Owner of the server",
                        },
                        {
                            "userId": "user2",
                            "group": "senior",
                            "comment": "Senior admin",
                        },
                    ],
                },
            ),
        ).get_admin_users()

    def test_commands_get_permanent_bans(self) -> None:
        SyncRconCommandsStub(
            "GetPermanentBans",
            2,
            response=json.dumps(
                {
                    "banList": [
                        {
                            "userId": "player1",
                            "userName": "Player One",
                            "timeOfBanning": "2023-10-01T12:00:00Z",
                            "durationHours": 0,
                            "banReason": "Cheating",
                            "adminName": "Admin1",
                        },
                        {
                            "userId": "player2",
                            "userName": "Player Two",
                            "timeOfBanning": "2023-10-02T12:00:00Z",
                            "durationHours": 0,
                            "banReason": "Toxic behavior",
                            "adminName": "Admin2",
                        },
                    ],
                },
            ),
        ).get_permanent_bans()

    def test_commands_get_temporary_bans(self) -> None:
        SyncRconCommandsStub(
            "GetTemporaryBans",
            2,
            response=json.dumps(
                {
                    "banList": [
                        {
                            "userId": "player3",
                            "userName": "Player Three",
                            "timeOfBanning": "2023-10-03T12:00:00Z",
                            "durationHours": 24,
                            "banReason": "AFK farming",
                            "adminName": "Admin3",
                        },
                        {
                            "userId": "player4",
                            "userName": "Player Four",
                            "timeOfBanning": "2023-10-04T12:00:00Z",
                            "durationHours": 48,
                            "banReason": "Inappropriate language",
                            "adminName": "Admin4",
                        },
                    ],
                },
            ),
        ).get_temporary_bans()

    def test_commands_disband_squad(self) -> None:
        team_id = 1
        squad_id = 2
        reason = "Disbanding for testing"
        SyncRconCommandsStub(
            "DisbandPlatoon",
            2,
            {
                "TeamIndex": team_id,
                "SquadIndex": squad_id,
                "Reason": reason,
            },
        ).disband_squad(team_id, squad_id, reason)

    def test_commands_force_team_switch(self) -> None:
        player_id = "player123"
        force_mode = ForceMode.IMMEDIATE
        response = SyncRconCommandsStub(
            "ForceTeamSwitch",
            2,
            {
                "PlayerId": player_id,
                "ForceMode": force_mode,
            },
        ).force_team_switch(player_id, force_mode)
        assert response is True

    def test_commands_set_team_switch_cooldown(self) -> None:
        minutes = 10
        SyncRconCommandsStub(
            "SetTeamSwitchCooldown",
            2,
            {"TeamSwitchTimer": minutes},
        ).set_team_switch_cooldown(minutes)

    def test_commands_set_max_queued_players(self) -> None:
        num = 20
        SyncRconCommandsStub(
            "SetMaxQueuedPlayers",
            2,
            {"MaxQueuedPlayers": num},
        ).set_max_queued_players(num)

    def test_commands_set_idle_kick_duration(self) -> None:
        minutes = 15
        SyncRconCommandsStub(
            "SetIdleKickDuration",
            2,
            {"IdleTimeoutMinutes": minutes},
        ).set_idle_kick_duration(minutes)

    def test_commands_message_all_players(self) -> None:
        message = "Hello all!"
        SyncRconCommandsStub(
            "SetWelcomeMessage",
            2,
            {"Message": message},
        ).set_welcome_message(message)

    def test_commands_get_player(self) -> None:
        player_id = "player123"

        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "player", "Value": player_id},
            json.dumps(
                {
                    "name": "Player1",
                    "clanTag": "ClanA",
                    "iD": player_id,
                    "platform": "steam",
                    "eosId": "1234567890",
                    "level": 25,
                    "team": 1,
                    "role": 1,
                    "platoon": "ABLE",
                    "loadout": "Combat Medic",
                    "kills": 150,
                    "deaths": 50,
                    "scoreData": {
                        "cOMBAT": 100,
                        "offense": 50,
                        "defense": 30,
                        "support": 20,
                    },
                    "worldPosition": {
                        "x": 1000,
                        "y": 2000,
                        "z": 300,
                    },
                },
            ),
        ).get_player(player_id)

    def test_commands_get_players(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "players", "Value": ""},
            json.dumps(
                {
                    "players": [
                        {
                            "name": "Player1",
                            "clanTag": "ClanA",
                            "iD": "123",
                            "platform": "steam",
                            "eosId": "1234567890",
                            "level": 25,
                            "team": 1,
                            "role": 1,
                            "platoon": "ABLE",
                            "loadout": "Combat Medic",
                            "kills": 150,
                            "deaths": 50,
                            "scoreData": {
                                "cOMBAT": 100,
                                "offense": 50,
                                "defense": 30,
                                "support": 20,
                            },
                            "worldPosition": {
                                "x": 1000,
                                "y": 2000,
                                "z": 300,
                            },
                        },
                    ],
                },
            ),
        ).get_players()

    def test_commands_get_map_rotation(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "maprotation", "Value": ""},
            json.dumps(
                {
                    "mAPS": [
                        {
                            "name": "Map1",
                            "gameMode": "Warfare",
                            "timeOfDay": "Day",
                            "iD": "map1",
                            "position": 0,
                        },
                        {
                            "name": "Map2",
                            "gameMode": "Offensive",
                            "timeOfDay": "Day",
                            "iD": "map2",
                            "position": 1,
                        },
                    ],
                },
            ),
        ).get_map_rotation()

    def test_commands_get_map_sequence(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "mapsequence", "Value": ""},
            json.dumps(
                {
                    "mAPS": [
                        {
                            "name": "Map1",
                            "gameMode": "Warfare",
                            "timeOfDay": "Day",
                            "iD": "map1",
                            "position": 0,
                        },
                        {
                            "name": "Map2",
                            "gameMode": "Offensive",
                            "timeOfDay": "Day",
                            "iD": "map2",
                            "position": 1,
                        },
                    ],
                },
            ),
        ).get_map_sequence()

    def test_commands_get_server_session(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "session", "Value": ""},
            json.dumps(
                {
                    "serverName": "My Server",
                    "mapName": "Map1",
                    "gameMode": "Warfare",
                    "remainingMatchTime": 0,
                    "matchTime": 6000,
                    "alliedScore": 2,
                    "axisScore": 2,
                    "playerCount": 98,
                    "alliedPlayerCount": 0,
                    "axisPlayerCount": 0,
                    "maxPlayerCount": 100,
                    "queueCount": 5,
                    "maxQueueCount": 6,
                    "vipQueueCount": 1,
                    "maxVipQueueCount": 2,
                },
            ),
        ).get_server_session()

    def test_commands_get_server_config(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "serverconfig", "Value": ""},
            json.dumps(
                {
                    "serverName": "My Server",
                    "buildNumber": "12345",
                    "buildRevision": "67890",
                    "supportedPlatforms": ["Steam", "WinGDK", "eos"],
                    "passwordProtected": False,
                },
            ),
        ).get_server_config()

    def test_commands_get_banned_words(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "bannedwords", "Value": ""},
            json.dumps(
                {
                    "bannedWords": ["garry", "blueberry"],
                },
            ),
        ).get_banned_words()

    def test_commands_get_vip_users(self) -> None:
        SyncRconCommandsStub(
            "GetServerInformation",
            2,
            {"Name": "vipplayers", "Value": ""},
            json.dumps(
                {
                    "vipPlayerIds": ["123", "456"],
                },
            ),
        ).get_vip_users()

    def test_commands_broadcast(self) -> None:
        message = "Broadcast message"
        SyncRconCommandsStub(
            "ServerBroadcast",
            2,
            {"Message": message},
        ).broadcast(message)

    def test_commands_set_high_ping_threshold(self) -> None:
        ms = 150
        SyncRconCommandsStub(
            "SetHighPingThreshold",
            2,
            {"HighPingThresholdMs": ms},
        ).set_high_ping_threshold(ms)

    def test_commands_get_command_details(self) -> None:
        command_id = "cmd1"
        SyncRconCommandsStub(
            "GetClientReferenceData",
            2,
            command_id,
            json.dumps(
                {
                    "name": command_id,
                    "text": "Command 1",
                    "description": "This is a test command.",
                    "dialogueParameters": [
                        {
                            "type": "Text",
                            "name": "param1",
                            "iD": "p1",
                            "displayMember": "",
                            "valueMember": "",
                        },
                        {
                            "type": "Number",
                            "name": "param2",
                            "iD": "p2",
                            "displayMember": "",
                            "valueMember": "",
                        },
                        {
                            "type": "Combo",
                            "name": "param3",
                            "iD": "p3",
                            "displayMember": "Option1,Option2",
                            "valueMember": "1,2",
                        },
                    ],
                },
            ),
        ).get_command_details(command_id)

    def test_commands_message_player(self) -> None:
        player_id = "pid"
        message = "Private message"
        SyncRconCommandsStub(
            "MessagePlayer",
            2,
            {"Message": message, "PlayerId": player_id},
        ).message_player(player_id, message)

    def test_commands_kill_player(self) -> None:
        player_id = "pid"
        reason = "Misconduct"
        result = SyncRconCommandsStub(
            "PunishPlayer",
            2,
            {"PlayerId": player_id, "Reason": reason},
        ).kill_player(player_id, reason)
        assert result is True

    def test_commands_kill_player_already_dead(self) -> None:
        player_id = "pid"
        reason = "Already dead"
        stub = SyncRconCommandsStub(
            "PunishPlayer",
            2,
            {"PlayerId": player_id, "Reason": reason},
            response=HLLCommandError(500, "Unable to perform request."),
        )

        result = stub.kill_player(player_id, reason)
        assert result is False

    def test_commands_kick_player(self) -> None:
        player_id = "pid"
        reason = "AFK"
        SyncRconCommandsStub(
            "KickPlayer",
            2,
            {"PlayerId": player_id, "Reason": reason},
        ).kick_player(player_id, reason)

    def test_commands_ban_player_permanent(self) -> None:
        player_id = "pid"
        reason = "Cheating"
        admin_name = "admin"
        SyncRconCommandsStub(
            "PermanentBanPlayer",
            2,
            {"PlayerId": player_id, "Reason": reason, "AdminName": admin_name},
        ).ban_player(player_id, reason, admin_name)

    def test_commands_ban_player_temporary(self) -> None:
        player_id = "pid"
        reason = "Toxic"
        admin_name = "admin"
        duration_hours = 12
        SyncRconCommandsStub(
            "TemporaryBanPlayer",
            2,
            {
                "PlayerId": player_id,
                "Duration": duration_hours,
                "Reason": reason,
                "AdminName": admin_name,
            },
        ).ban_player(player_id, reason, admin_name, duration_hours=duration_hours)

    def test_commands_remove_temp_ban(self) -> None:
        player_id = "pid"
        SyncRconCommandsStub(
            "RemoveTemporaryBan",
            2,
            {"PlayerId": player_id},
        ).remove_temporary_ban(player_id)

    def test_commands_remove_permanent_ban(self) -> None:
        player_id = "pid"
        SyncRconCommandsStub(
            "RemovePermanentBan",
            2,
            {"PlayerId": player_id},
        ).remove_permanent_ban(player_id)

    def test_commands_remove_ban(self) -> None:
        commands = mock.Mock(spec=SyncRconCommands)
        commands.remove_ban = partial(SyncRconCommands.unban_player, commands)
        commands.remove_ban("pid")
        commands.remove_temporary_ban.assert_called_once_with("pid")
        commands.remove_permanent_ban.assert_called_once_with("pid")

    def test_commands_remove_player_from_squad(self) -> None:
        player_id = "pid"
        reason = "Squad disbanded"
        SyncRconCommandsStub(
            "RemovePlayerFromPlatoon",
            2,
            {
                "PlayerId": player_id,
                "Reason": reason,
            },
        ).remove_player_from_squad(player_id, reason)

    def test_commands_set_auto_balance_enabled(self) -> None:
        SyncRconCommandsStub(
            "SetAutoBalanceEnabled",
            2,
            {"Enable": True},
        ).set_auto_balance_enabled(enabled=True)

    def test_commands_set_auto_balance_threshold(self) -> None:
        threshold = 5
        SyncRconCommandsStub(
            "SetAutoBalanceThreshold",
            2,
            {"AutoBalanceThreshold": threshold},
        ).set_auto_balance_threshold(threshold)

    def test_commands_set_vote_kick_enabled(self) -> None:
        SyncRconCommandsStub(
            "SetVoteKickEnabled",
            2,
            {"Enable": True},
        ).set_vote_kick_enabled(enabled=True)

    def test_commands_reset_vote_kick_thresholds(self) -> None:
        SyncRconCommandsStub(
            "ResetVoteKickThreshold",
            2,
        ).reset_vote_kick_thresholds()

    def test_commands_set_vote_kick_thresholds(self) -> None:
        thresholds = [(10, 2), (20, 3)]
        SyncRconCommandsStub(
            "SetVoteKickThreshold",
            2,
            {"ThresholdValue": "10,2,20,3"},
        ).set_vote_kick_thresholds(thresholds)

    def test_commands_add_banned_words(self) -> None:
        words = ["badword1", "badword2"]
        SyncRconCommandsStub(
            "AddBannedWords",
            2,
            {"Words": ",".join(words)},
        ).add_banned_words(words)

    def test_commands_remove_banned_words(self) -> None:
        words = ["badword1", "badword2"]
        SyncRconCommandsStub(
            "RemoveBannedWords",
            2,
            {"Words": ",".join(words)},
        ).remove_banned_words(words)

    def test_commands_add_vip(self) -> None:
        player_id = "vip123"
        description = "desc"
        SyncRconCommandsStub(
            "AddVip",
            2,
            {"PlayerId": player_id, "Comment": description},
        ).add_vip(player_id, description)

    def test_commands_remove_vip(self) -> None:
        player_id = "vip123"
        SyncRconCommandsStub(
            "RemoveVip",
            2,
            {"PlayerId": player_id},
        ).remove_vip(player_id)

    def test_commands_set_num_vip_slots(self) -> None:
        num_slots = 5
        SyncRconCommandsStub(
            "SetVipSlotCount",
            2,
            {"VipSlotCount": num_slots},
        ).set_num_vip_slots(num_slots)

    def test_commands_set_match_timer(self) -> None:
        game_mode: Literal["Warfare"] = "Warfare"
        minutes = 30
        SyncRconCommandsStub(
            "SetMatchTimer",
            2,
            {"GameMode": game_mode, "MatchLength": minutes},
        ).set_match_timer(game_mode, minutes)

    def test_commands_remove_match_timer(self) -> None:
        game_mode: Literal["Warfare"] = "Warfare"
        SyncRconCommandsStub(
            "RemoveMatchTimer",
            2,
            {"GameMode": game_mode},
        ).reset_match_timer(game_mode)

    def test_commands_set_warmup_timer(self) -> None:
        game_mode: Literal["Warfare"] = "Warfare"
        minutes = 5
        SyncRconCommandsStub(
            "SetWarmupTimer",
            2,
            {"GameMode": game_mode, "WarmupLength": minutes},
        ).set_warmup_timer(game_mode, minutes)

    def test_commands_remove_warmup_timer(self) -> None:
        game_mode: Literal["Warfare"] = "Warfare"
        SyncRconCommandsStub(
            "RemoveWarmupTimer",
            2,
            {"GameMode": game_mode},
        ).remove_warmup_timer(game_mode)

    def test_commands_set_dynamic_weather_enabled(self) -> None:
        map_id = "map123"
        enabled = True
        SyncRconCommandsStub(
            "SetDynamicWeatherEnabled",
            2,
            {"MapId": map_id, "Enable": enabled},
        ).set_dynamic_weather_enabled(map_id, enabled=enabled)



import asyncio
import logging
import os
from collections.abc import AsyncGenerator

import pytest
import pytest_asyncio
from hllrcon.rcon import Rcon
from hllrcon.responses import (
    GetCommandDetailsResponse,
    GetCommandsResponse,
    GetMapRotationResponseEntry,
    GetPlayerResponse,
    GetServerConfigResponse,
    GetServerSessionResponse,
)
from pydantic import TypeAdapter

HLL_HOST = os.getenv("HLL_HOST")
HLL_PORT = os.getenv("HLL_PORT")
HLL_PASSWORD = os.getenv("HLL_PASSWORD")

if not HLL_HOST or not HLL_PORT or not HLL_PASSWORD:
    pytest.skip("HLL environment variables are not set", allow_module_level=True)

pytestmark = pytest.mark.asyncio


@pytest_asyncio.fixture(scope="function")
async def rcon() -> AsyncGenerator[Rcon]:
    rcon = Rcon(
        host=str(HLL_HOST),
        port=int(HLL_PORT or ""),
        password=str(HLL_PASSWORD),
    )
    await asyncio.sleep(0.1)
    async with rcon.connect():
        yield rcon


@pytest_asyncio.fixture
async def players(rcon: Rcon) -> list[GetPlayerResponse]:
    players = await rcon.get_players()

    if not players.players:
        pytest.skip("No players found on the server")

    return players.players


@pytest_asyncio.fixture
async def rotation(rcon: Rcon) -> list[GetMapRotationResponseEntry]:
    rotation = await rcon.get_map_rotation()
    return rotation.maps


@pytest_asyncio.fixture
async def sequence(rcon: Rcon) -> list[GetMapRotationResponseEntry]:
    rotation = await rcon.get_map_sequence()
    return rotation.maps


@pytest_asyncio.fixture
async def server_config(rcon: Rcon) -> GetServerConfigResponse:
    return await rcon.get_server_config()


@pytest_asyncio.fixture
async def server_session(rcon: Rcon) -> GetServerSessionResponse:
    return await rcon.get_server_session()


class TestIntegratedServer:
    @pytest.fixture(autouse=True)
    def setup_module(self, caplog: pytest.LogCaptureFixture) -> None:
        caplog.set_level(logging.DEBUG)

    async def test_validate_player_data(self, players: list[GetPlayerResponse]) -> None:
        pass

    async def test_validate_rotation_data(
        self,
        rotation: list[GetMapRotationResponseEntry],
    ) -> None:
        pass

    async def test_validate_server_session_data(
        self,
        server_session: GetServerSessionResponse,
    ) -> None:
        pass

    async def test_validate_server_config_data(
        self,
        server_config: GetServerConfigResponse,
    ) -> None:
        pass

    async def test_validate_client_reference_data(self, rcon: Rcon) -> None:
        commands = await rcon.get_commands()
        TypeAdapter(GetCommandsResponse).validate_python(commands)

        details = await asyncio.gather(
            *(rcon.get_command_details(command.id) for command in commands.entries),
        )
        TypeAdapter(list[GetCommandDetailsResponse]).validate_python(details)

    async def test_kill_missing_player_returns_false(self, rcon: Rcon) -> None:
        result = await rcon.kill_player("1234567890", "Test reason")
        assert result is False

    async def test_kick_missing_player_returns_false(self, rcon: Rcon) -> None:
        result = await rcon.kick_player("1234567890", "Test reason")
        assert result is False

    async def test_modify_sequence(
        self,
        rcon: Rcon,
        sequence: list[GetMapRotationResponseEntry],
    ) -> None:
        index = len(sequence)

        await rcon.add_map_to_sequence("foy_warfare", index)

        new_sequence = await rcon.get_map_sequence()
        assert new_sequence.maps == [
            *sequence,
            GetMapRotationResponseEntry(
                name="FOY",
                game_mode_name="Warfare",
                time_of_day="Day",
                id="/Game/Maps/foy_warfare",
                position=index,
            ),
        ]

        await rcon.remove_map_from_sequence(index)

        new_sequence = await rcon.get_map_sequence()
        assert new_sequence.maps == sequence



import asyncio
import base64
import binascii
import itertools
import json
from unittest.mock import Mock

import pytest
import pytest_asyncio
from hllrcon.exceptions import (
    HLLAuthError,
    HLLConnectionError,
    HLLConnectionLostError,
    HLLConnectionRefusedError,
    HLLMessageError,
)
from hllrcon.protocol.protocol import RconProtocol
from hllrcon.protocol.request import RconRequest
from hllrcon.protocol.response import RconResponse, RconResponseStatus
from pytest_mock import MockerFixture


@pytest.fixture
def transport(mocker: MockerFixture) -> asyncio.Transport:
    transport: Mock = mocker.Mock(spec=asyncio.Transport)
    transport.is_closing.return_value = False
    return transport


@pytest.fixture
def load_constants(request: pytest.FixtureRequest, mocker: MockerFixture) -> None:
    if marker := request.node.get_closest_marker("do_allow_concurrent_requests"):
        mocker.patch(
            "hllrcon.protocol.protocol.DO_ALLOW_CONCURRENT_REQUESTS",
            bool(marker.args[0]) if marker.args else True,
        )

    if marker := request.node.get_closest_marker("do_pop_v1_xorkey"):
        mocker.patch(
            "hllrcon.protocol.protocol.DO_POP_V1_XORKEY",
            bool(marker.args[0]) if marker.args else True,
        )

    if marker := request.node.get_closest_marker("do_use_request_headers"):
        mocker.patch(
            "hllrcon.protocol.protocol.DO_USE_REQUEST_HEADERS",
            bool(marker.args[0]) if marker.args else True,
        )

    if marker := request.node.get_closest_marker("do_xor_responses"):
        mocker.patch(
            "hllrcon.protocol.protocol.DO_XOR_RESPONSES",
            bool(marker.args[0]) if marker.args else False,
        )


@pytest_asyncio.fixture
async def protocol(
    load_constants: None,  # noqa: ARG001
    transport: asyncio.Transport,
    mocker: MockerFixture,
) -> asyncio.Protocol:
    protocol = RconProtocol(asyncio.get_running_loop(), timeout=1.0)
    protocol.connection_made(transport)
    mocker.patch.object(
        RconRequest,
        "_RconRequest__request_id_counter",
        itertools.count(start=1),
    )
    return protocol


def test_is_connected(protocol: RconProtocol, transport: Mock) -> None:
    assert protocol.is_connected() is True

    transport.is_closing.return_value = True
    assert protocol.is_connected() is False

    protocol._transport = None
    assert protocol.is_connected() is False


@pytest.mark.asyncio
async def test_connect(
    mocker: MockerFixture,
    protocol: RconProtocol,
    transport: Mock,
) -> None:
    mock_create_connection = mocker.patch.object(
        asyncio.get_running_loop(),
        "create_connection",
        return_value=(transport, protocol),
    )

    host = "localhost"
    port = 1234
    password = "password"

    mock_create_connection.side_effect = TimeoutError
    with pytest.raises(
        HLLConnectionError,
        match=f"Address {host} could not be resolved",
    ):
        await protocol.connect(host, port, password)

    mock_create_connection.side_effect = ConnectionRefusedError
    with pytest.raises(
        HLLConnectionRefusedError,
        match=f"The server refused connection over port {port}",
    ):
        await protocol.connect(host, port, password)

    mock_create_connection.side_effect = None
    authenticate = mocker.patch.object(protocol, "authenticate")

    authenticate.side_effect = HLLAuthError
    with pytest.raises(HLLAuthError):
        await protocol.connect(host, port, password)
    assert protocol._transport is None

    authenticate.side_effect = None
    result = await protocol.connect(host, port, password)
    assert result is protocol


def test_disconnect(protocol: RconProtocol, transport: Mock) -> None:
    protocol.disconnect()

    transport.close.assert_called_once()
    assert protocol._transport is None

    protocol.disconnect()


def test_connection_made(protocol: RconProtocol) -> None:
    with pytest.raises(
        TypeError,
        match="Transport must be an instance of asyncio.Transport",
    ):
        protocol.connection_made(Mock(spec=asyncio.BaseTransport))


def test_data_received(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    mocker.patch.object(protocol, "_read_from_buffer")

    data1 = b"Some data after xorkey"
    protocol.data_received(data1)

    data2 = b"Even more data"
    protocol.data_received(data2)

    assert protocol._buffer == data1 + data2


def test_read_from_buffer_too_small(
    protocol: RconProtocol,
) -> None:
    data = b"\x01\x02\x03\x04\x05\x06\x07"

    protocol._buffer = data
    protocol._read_from_buffer()
    assert protocol._buffer == data


def test_read_from_buffer_incomplete_packet(
    protocol: RconProtocol,
) -> None:
    data = b"\x01\x00\x00\x00\x05\x00\x00\x00Hell"

    protocol._buffer = data
    protocol._read_from_buffer()
    assert protocol._buffer == data


def test_read_from_buffer_exactly_one_packet(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    mock_unpack = mocker.patch(
        "hllrcon.protocol.protocol.RconResponse.unpack",
        autospec=True,
    )
    data = b"\x01\x00\x00\x00\x05\x00\x00\x00Hello"

    waiter: asyncio.Future[RconResponse] = asyncio.Future()
    protocol._waiters[1] = waiter

    protocol._buffer = data
    protocol._read_from_buffer()
    assert protocol._buffer == b""
    assert waiter.result()
    mock_unpack.assert_called_once_with(1, b"Hello")


def test_read_from_buffer_exactly_one_packet_missing_waiter(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    mock_unpack = mocker.patch(
        "hllrcon.protocol.protocol.RconResponse.unpack",
        autospec=True,
    )
    data = b"\x01\x00\x00\x00\x05\x00\x00\x00Hello"

    protocol._buffer = data
    protocol._read_from_buffer()
    assert protocol._buffer == b""
    mock_unpack.assert_called_once_with(1, b"Hello")


def test_read_from_buffer_more_than_one_packet(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    mock_unpack = mocker.patch(
        "hllrcon.protocol.protocol.RconResponse.unpack",
        autospec=True,
    )
    data = (
        b"\x01\x00\x00\x00\x05\x00\x00\x00Hello"
        b"\x02\x00\x00\x00\x05\x00\x00\x00World"
        b"\x00\x00"
    )

    waiter1: asyncio.Future[RconResponse] = asyncio.Future()
    waiter2: asyncio.Future[RconResponse] = asyncio.Future()
    protocol._waiters[1] = waiter1
    protocol._waiters[2] = waiter2

    protocol._buffer = data
    protocol._read_from_buffer()
    assert protocol._buffer == b"\x00\x00"
    assert waiter1.result()
    assert waiter2.result()
    assert mock_unpack.call_count == 2
    mock_unpack.assert_any_call(1, b"Hello")
    mock_unpack.assert_called_with(2, b"World")


def test_connection_lost_use_request_headers(
    protocol: RconProtocol,
) -> None:
    waiters: dict[int, asyncio.Future[RconResponse]] = {
        1: asyncio.Future(),
        2: asyncio.Future(),
    }
    protocol._waiters = waiters.copy()

    protocol.connection_lost(None)

    assert not protocol.is_connected()
    assert not protocol._waiters
    assert waiters[1].cancelled()
    assert waiters[2].cancelled()


def test_connection_lost_with_exception(
    protocol: RconProtocol,
) -> None:
    waiters: dict[int, asyncio.Future[RconResponse]] = {
        1: asyncio.Future(),
        2: asyncio.Future(),
    }
    protocol._waiters = waiters.copy()

    response = RconResponse(
        request_id=1,
        command="command",
        version=1,
        status_code=RconResponseStatus.OK,
        status_message="OK",
        content_body="foo",
    )
    waiters[1].set_result(response)

    protocol.connection_lost(OSError("Connection error"))

    assert not protocol.is_connected()
    assert not protocol._waiters
    assert waiters[1].result() == response
    assert isinstance(waiters[2].exception(), HLLConnectionLostError)


def test_connection_lost_invokes_callback(
    protocol: RconProtocol,
    mocker: MockerFixture,
) -> None:
    connection_lost_callback = mocker.Mock()
    protocol.on_connection_lost = connection_lost_callback
    protocol.connection_lost(None)
    connection_lost_callback.assert_called_once_with(None)


def test_connection_lost_callback_invocation_failure(
    protocol: RconProtocol,
    mocker: MockerFixture,
) -> None:
    connection_lost_callback = mocker.Mock(side_effect=RuntimeError("Callback failed"))
    protocol.on_connection_lost = connection_lost_callback
    protocol.connection_lost(None)
    connection_lost_callback.assert_called_once_with(None)


def test_xor_single_byte_key(protocol: RconProtocol) -> None:
    protocol.xorkey = b"\x01"
    msg = b"\x00\x01\x02"
    # Each byte XOR 0x01
    expected = bytes([b ^ 0x01 for b in msg])
    assert protocol._xor(msg) == expected


def test_xor_multi_byte_key(protocol: RconProtocol) -> None:
    protocol.xorkey = b"\x01\x02\x03"
    msg = b"\x10\x20\x30\x40\x50\x60"
    # XOR with repeating key
    expected = bytes(
        [
            0x10 ^ 0x01,  # 0
            0x20 ^ 0x02,  # 1
            0x30 ^ 0x03,  # 2
            0x40 ^ 0x01,  # 3
            0x50 ^ 0x02,  # 4
            0x60 ^ 0x03,  # 5
        ],
    )
    assert protocol._xor(msg) == expected


def test_xor_offset(protocol: RconProtocol) -> None:
    protocol.xorkey = b"\x01\x02\x03"
    msg = b"\x10\x20\x30"
    # With offset=1, key index starts at 1
    expected = bytes(
        [
            0x10 ^ 0x02,  # (0+1)%3 = 1
            0x20 ^ 0x03,  # (1+1)%3 = 2
            0x30 ^ 0x01,  # (2+1)%3 = 0
        ],
    )
    assert protocol._xor(msg, offset=1) == expected


def test_xor_roundtrip(protocol: RconProtocol) -> None:
    protocol.xorkey = b"\x0a\x0b\x0c"
    msg = b"SecretMessage"
    encrypted = protocol._xor(msg)
    # XOR again should recover original
    decrypted = protocol._xor(encrypted)
    assert decrypted == msg


def test_xor_length_mismatch_raises(
    protocol: RconProtocol,
    mocker: MockerFixture,
) -> None:
    protocol.xorkey = b"\x01"
    msg = b"\x01\x02"
    # Patch array.array to return wrong length
    mock_array = mocker.patch("array.array")
    mock_array.return_value.tobytes.return_value = b"\x00"
    with pytest.raises(
        ValueError,
        match="XOR operation resulted in a different length",
    ):
        protocol._xor(msg)


def test_xor_warns_on_length_mismatch(
    protocol: RconProtocol,
    mocker: MockerFixture,
) -> None:
    protocol.xorkey = b"\x01"
    msg = b"\x01\x02"
    # Patch array.array to return wrong length and check logger.warning called
    mock_array = mocker.patch("array.array")
    mock_array.return_value.tobytes.return_value = b"\x00"
    mock_logger = mocker.patch.object(protocol.logger, "warning")
    with pytest.raises(
        ValueError,
        match="XOR operation resulted in a different length",
    ):
        protocol._xor(msg)
    mock_logger.assert_called_with(
        "XOR operation resulted in a different length: %s != %s",
        1,
        2,
    )


@pytest.mark.asyncio
async def test_execute_no_connection(
    protocol: RconProtocol,
) -> None:
    protocol.connection_lost(None)
    with pytest.raises(HLLConnectionError, match="Connection is closed"):
        await protocol.execute("command", 1, "body")


def make_response(request_id: int, message: str) -> bytes:
    body = {
        "name": "command",
        "version": 1,
        "statusCode": 200,
        "statusMessage": "OK",
        "contentBody": message,
    }
    body_encoded = json.dumps(body).encode("utf-8")
    return (
        request_id.to_bytes(4, "little")
        + len(body_encoded).to_bytes(4, "little")
        + body_encoded
    )


@pytest.mark.asyncio
async def test_execute(
    protocol: RconProtocol,
    transport: Mock,
) -> None:
    asyncio.get_event_loop().call_later(
        0.1,
        protocol.data_received,
        make_response(0, "response"),
    )
    response = await protocol.execute("command", 1, "body")
    assert response.content_body == "response"
    transport.write.assert_called_once()


@pytest.mark.asyncio
async def test_execute_with_timeout(
    protocol: RconProtocol,
    transport: Mock,
) -> None:
    protocol.timeout = 0.1
    with pytest.raises(TimeoutError):
        await protocol.execute("command", 1, "body")
    assert not protocol._waiters
    transport.write.assert_called_once()


@pytest.mark.asyncio
async def test_execute_concurrently(
    protocol: RconProtocol,
    transport: Mock,
) -> None:
    protocol.timeout = 1.0
    asyncio.get_running_loop().call_later(
        0.5,
        protocol.data_received,
        make_response(1, "response2") + make_response(0, "response1"),
    )
    responses = await asyncio.gather(
        protocol.execute("command1", 1, "body1"),
        protocol.execute("command2", 2, "body2"),
    )
    assert responses[0].content_body == "response1"
    assert responses[1].content_body == "response2"
    assert transport.write.call_count == 2


@pytest.mark.asyncio
async def test_authenticate_success(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    # Mock execute to return fake responses for ServerConnect and Login
    xorkey_b64 = base64.b64encode(b"keybytes").decode()
    xorkey_response = mocker.Mock()
    xorkey_response.content_body = xorkey_b64
    xorkey_response.raise_for_status = mocker.Mock()

    auth_token_response = mocker.Mock()
    auth_token_response.content_body = "token123"
    auth_token_response.raise_for_status = mocker.Mock()

    execute = mocker.patch.object(
        protocol,
        "execute",
        side_effect=[xorkey_response, auth_token_response],
    )

    await protocol.authenticate("password")

    execute.assert_has_calls(
        [
            mocker.call("ServerConnect", 2, ""),
            mocker.call("Login", 2, "password"),
        ],
    )
    xorkey_response.raise_for_status.assert_called_once()
    auth_token_response.raise_for_status.assert_called_once()
    assert protocol.xorkey == b"keybytes"
    assert protocol.auth_token == "token123"


@pytest.mark.asyncio
async def test_authenticate_serverconnect_not_string(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    # Mock execute to return a non-string content_body for ServerConnect
    xorkey_response = mocker.Mock()
    xorkey_response.content_body = 12345  # Not a string
    xorkey_response.raise_for_status = mocker.Mock()

    execute = mocker.patch.object(protocol, "execute", side_effect=[xorkey_response])

    with pytest.raises(
        HLLMessageError,
        match="ServerConnect response content_body is not a string",
    ):
        await protocol.authenticate("password")

    execute.assert_called_once_with("ServerConnect", 2, "")


@pytest.mark.asyncio
async def test_authenticate_serverconnect_raises_for_status(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    # Mock execute to raise for status on ServerConnect
    xorkey_response = mocker.Mock()
    xorkey_response.content_body = "ignored"
    xorkey_response.raise_for_status = mocker.Mock(side_effect=Exception("fail"))

    execute = mocker.patch.object(protocol, "execute", side_effect=[xorkey_response])

    with pytest.raises(Exception, match="fail"):
        await protocol.authenticate("password")

    execute.assert_called_once_with("ServerConnect", 2, "")


@pytest.mark.asyncio
async def test_authenticate_login_raises_for_status(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    # Mock execute to raise for status on Login
    xorkey_b64 = base64.b64encode(b"keybytes").decode()
    xorkey_response = mocker.Mock()
    xorkey_response.content_body = xorkey_b64
    xorkey_response.raise_for_status = mocker.Mock()

    auth_token_response = mocker.Mock()
    auth_token_response.content_body = "token123"
    auth_token_response.raise_for_status = mocker.Mock(
        side_effect=Exception("loginfail"),
    )

    mocker.patch.object(
        protocol,
        "execute",
        side_effect=[xorkey_response, auth_token_response],
    )

    with pytest.raises(Exception, match="loginfail"):
        await protocol.authenticate("password")

    assert protocol.xorkey == b"keybytes"
    # auth_token should not be set on failure
    assert protocol.auth_token is None


@pytest.mark.asyncio
async def test_authenticate_xorkey_base64_decode_error(
    mocker: MockerFixture,
    protocol: RconProtocol,
) -> None:
    # Mock execute to return invalid base64 for xorkey
    xorkey_response = mocker.Mock()
    xorkey_response.content_body = "!!!notbase64!!!"
    xorkey_response.raise_for_status = mocker.Mock()

    mocker.patch.object(protocol, "execute", side_effect=[xorkey_response])

    with pytest.raises(binascii.Error):
        await protocol.authenticate("password")
    # xorkey should not be set
    assert protocol.xorkey is None
